// naumann@stce.rwth-aachen.de

#include "f.hpp"

#include <vector>
#include <cmath>
#include <limits>

template<typename T, typename PT>
void gradient(size_t ncp, size_t ncs, const PT &eps, T &x, std::vector<T> &p, const std::vector<std::vector<PT>> &dW, std::vector<T> &dxdp) {
  size_t n=p.size();
  for (size_t i=0;i<n;i++) {
    T h= (p[i]==0) 
        ? sqrt(sqrt(std::numeric_limits<T>::epsilon())) 
        : sqrt(sqrt(std::numeric_limits<T>::epsilon()))*fabs(p[i]);
    T x_p=x; p[i]+=h; f(ncp,ncs,eps,x_p,p,dW);
    T x_m=x; p[i]-=2*h; f(ncp,ncs,eps,x_m,p,dW); p[i]+=h;
    dxdp[i]=(x_p-x_m)/(2*h); 
  }
  f(ncp,ncs,eps,x,p,dW);
}  

template<typename T, typename PT>
void hessian(size_t ncp, size_t ncs, const PT &eps, T &x, std::vector<T> &p, const std::vector<std::vector<PT>> &dW, std::vector<T> &dxdp, std::vector<std::vector<T>>& ddxdpp) {
  size_t n=p.size();
  std::vector<T> dxdp_p(n), dxdp_m(n);
  for (size_t i=0;i<n;i++) {
    T h= (p[i]==0) 
        ? sqrt(sqrt(std::numeric_limits<T>::epsilon())) 
        : sqrt(sqrt(std::numeric_limits<T>::epsilon()))*fabs(p[i]);
    T x_p=x; p[i]+=h; gradient(ncp,ncs,eps,x_p,p,dW,dxdp_p);
    T x_m=x; p[i]-=2*h; gradient(ncp,ncs,eps,x_m,p,dW,dxdp_m); p[i]+=h;
    dxdp[i]=(x_p-x_m)/(2*h); 
    for (size_t j=0;j<n;j++) 
      ddxdpp[j][i]=(dxdp_p[j]-dxdp_m[j])/(2*h); 
  }
  f(ncp,ncs,eps,x,p,dW);
}  

#include "h_main.hpp"
